/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include <pixman.h>
#include <cmath>

#include "composite-test.h"
#include "composite-canvas-pixman.h"

CompositeTestPixman::CompositeTestPixman() :
    CompositeTest("pixman", "pixman")
{
    options_["filter"] = CompositeTest::Option("filter", "bilinear",
                                               "Filter to use [nearest, bilinear]");
}

void
CompositeTestPixman::init()
{
    pixman_color_t half_transparent = {0x7fff, 0x7fff, 0x7fff, 0x7fff};
    half_transparent_mask_ = pixman_image_create_solid_fill(&half_transparent);
}

void
CompositeTestPixman::deinit()
{
    if (half_transparent_mask_)
        pixman_image_unref(half_transparent_mask_);
}

void
CompositeTestPixman::prepare_for_run()
{
    if (options_["filter"].value == "bilinear")
        pixman_filter_ = PIXMAN_FILTER_BILINEAR;
    else
        pixman_filter_ = PIXMAN_FILTER_NEAREST;
}

void
CompositeTestPixman::draw(std::list<CompositeWindow *> &window_list)
{
    CompositeCanvasPixman *canvas_pixman =
        static_cast<CompositeCanvasPixman *>(canvas_);

    pixman_image_t *dest_image = canvas_pixman->pixman_image();
    int dest_width = pixman_image_get_width(dest_image);
    int dest_height = pixman_image_get_height(dest_image);

    /* Clear the screen with color matching GL {0.1, 0.1, 0.3, 1.0} */
    pixman_color_t clear = {6554, 6554, 19661, 65535};

    pixman_box32_t box;
    box.x1 = box.y1 = 0;
    box.x2 = dest_width > 0 ? dest_width - 1 : 0;
    box.y2 = dest_height > 0 ? dest_height - 1 : 0;

    pixman_image_fill_boxes(PIXMAN_OP_SRC, dest_image, &clear, 1, &box);

    /* Find out how many windows are visible and calculate the angular step */
    int visible_windows(num_visible_windows(window_list));
    float angular_step(2 * M_PI / visible_windows);

    /* Draw the windows in a circle using the calculated angular step */
    int i(0);
    for(std::list<CompositeWindow*>::iterator iter = window_list.begin();
        iter != window_list.end(); ++iter)
    {
        CompositeWindow *comp_win = *iter;
        pixman_image_t *pixman_image = reinterpret_cast<pixman_image_t *>(comp_win->get_texture().p);
        if (pixman_image) {
            int width = pixman_image_get_width(pixman_image);
            int height = pixman_image_get_height(pixman_image);

            /* Calculate the new position of the composited window center */
            int x0 = dest_width / 2 + cos(i * angular_step) * dest_width / 4;
            int y0 = dest_height / 2 + sin(i * angular_step) * dest_height / 4;

            /* Calculate the new position of the upper left corner */
            x0 -= dest_width / 4;
            y0 -= dest_height / 4;

            /* Scale the window to fit in dest_width/2 and dest_height/2 */
            pixman_transform_t xform;

            pixman_image_set_filter(pixman_image, pixman_filter_, NULL, 0);
            pixman_transform_init_scale(&xform,
                                        pixman_double_to_fixed(2.0 * width / dest_width),
                                        pixman_double_to_fixed(2.0 * height / dest_height));
            pixman_image_set_transform(pixman_image, &xform);

            /* Draw the image to the canvas image */
            pixman_image_composite32(PIXMAN_OP_OVER, pixman_image,
                                     half_transparent_mask_, dest_image,
                                     0, 0, 0, 0, x0, y0,
                                     dest_width / 2, dest_height / 2);

            i++;
        }
    }
}


