/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include <X11/Xlib.h>

#define GLX_GLXEXT_PROTOTYPES
#include <GL/gl.h>
#include <GL/glx.h>
#include <GL/glxext.h>

#include "composite-window-glxpixmap.h"
#include "log.h"

extern PFNGLXBINDTEXIMAGEEXTPROC glXBindTexImageEXT_;
extern PFNGLXRELEASETEXIMAGEEXTPROC glXReleaseTexImageEXT_;

void
CompositeWindowGLXPixmap::release_tfp()
{
    if (glx_pixmap_)
    {
        glXReleaseTexImageEXT_(xdpy_, glx_pixmap_, GLX_FRONT_LEFT_EXT);
        glXDestroyPixmap(xdpy_, glx_pixmap_);
        glx_pixmap_ = 0;
        glx_pixmap_bound_ = false;
    }
}

CompositeWindowGLXPixmap::~CompositeWindowGLXPixmap()
{
    release_tfp();
}

void
CompositeWindowGLXPixmap::update_texture_from_pixmap(bool recreate)
{
    const int pixmap_attribs[] = {
        GLX_TEXTURE_TARGET_EXT, GLX_TEXTURE_2D_EXT,
        GLX_TEXTURE_FORMAT_EXT, GLX_TEXTURE_FORMAT_RGBA_EXT,
        None
    };

    Profiler::ScopedSamples scoped_tfp(CompositeWindow::get_profiler_tfp_pair());

    if (recreate) {
        release_tfp();
        glx_pixmap_ = glXCreatePixmap(xdpy_, glx_fbconfig_, pix_,
                                      pixmap_attribs);

        Log::debug("    => GLXPixmap <= %p\n", glx_pixmap_);
    }

    /* Update texture with new data */
    glXWaitX();

    /*
     * According to the spec, proper TFP usage is:
     *
     * while(1) {
     *     glXWaitX();
     *     glXBindTexImageEXT();
     *     <Render using texture>
     *     glXReleaseTexImageEXT();
     * }
     *
     * we are using:
     *
     * while(1) {
     *     glXWaitX();
     *     if (bound) glXReleaseTexImageEXT();
     *     glxBindTexImageEXT();
     *     <Render using texture>
     * }
     *
     * The only potential problem here is that rendering to a pixmap bound to a
     * texture leaves the contents of the texture in an undefined state
     * (according to the spec).  We are not performing any explicit pixmap
     * rendering, but we do get implicit pixmap changes due to window
     * redirection. It is not clear if these changes should affect the texture.
     *
     * In any case, both compiz and cogl/clutter handle TFP this
     * way, so it seems to be safe for now for all drivers.
     */
    glBindTexture(GL_TEXTURE_2D, tex_);

    if (glx_pixmap_bound_)
        glXReleaseTexImageEXT_(xdpy_, glx_pixmap_, GLX_FRONT_LEFT_EXT);

    glXBindTexImageEXT_(xdpy_, glx_pixmap_, GLX_FRONT_LEFT_EXT, NULL);

    glx_pixmap_bound_ = true;
}
