/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <cstdlib>

#include "gl-headers.h"

#include "composite-window-pixman.h"
#include "log.h"

CompositeWindowPixman::~CompositeWindowPixman()
{
    release_resources();
}

void
CompositeWindowPixman::release_resources()
{
    if (ximage_)
        XDestroyImage(ximage_);

    if (pixman_image_)
        pixman_image_unref(pixman_image_);

    if (shm_info_.shmaddr != reinterpret_cast<char *>(-1)) {
        shmdt(shm_info_.shmaddr);
        XShmDetach(xdpy_, &shm_info_);
    }

    ximage_ = 0;
    pixman_image_ = 0;
    shm_info_.shmaddr = reinterpret_cast<char *>(-1);
}

bool
CompositeWindowPixman::have_xshm()
{
    int ignore, major, minor;
    Bool pixmaps;
    bool have_xshm = false;
        
    if (XQueryExtension(xdpy_, "MIT-SHM", &ignore, &ignore, &ignore)
        && XShmQueryVersion(xdpy_, &major, &minor, &pixmaps) == True)
    {
        have_xshm = true;
    }

    return have_xshm;
}

bool
CompositeWindowPixman::ensure_ximage_shm()
{
    if (ximage_)
        return true;

    XWindowAttributes win_attr;

    if (!XGetWindowAttributes(xdpy_, win_, &win_attr)) {
        Log::error("XGetWindowAttributes failed!\n");
        return false;
    }

    ximage_ = XShmCreateImage(xdpy_, win_attr.visual, win_attr.depth,
                              ZPixmap, 0, &shm_info_, width_, height_);
    if (!ximage_) {
        Log::error("XCreateImage failed!\n");
        return false;
    }
    
    /* Get shared memory from the system */
    shm_info_.shmid = shmget(IPC_PRIVATE, ximage_->bytes_per_line * ximage_->height,
                             IPC_CREAT | 0777);
    if (shm_info_.shmid < 0) {
        Log::error("shmget failed!\n");
        release_resources();
        return false;
    }

    /* Attach shared memory to process address space and set it as RW */
    shm_info_.readOnly = False;
    shm_info_.shmaddr = (char *)shmat(shm_info_.shmid, 0, 0);
    if (shm_info_.shmaddr == reinterpret_cast<char *>(-1)) {
        Log::error("shmat failed!\n");
        release_resources();
        return false;
    }

    /* Set shared memory as the pixel data for the XImage*/
    ximage_->data = shm_info_.shmaddr;
        
    /* Attach shared memory to X Server */
    if (!XShmAttach(xdpy_, &shm_info_)) {
        Log::error("XShmAttach failed!\n");
        release_resources();
        return false;
    }

    return true;
}

bool
CompositeWindowPixman::ensure_ximage()
{
    if (ximage_)
        return true;

    XWindowAttributes win_attr;

    if (!XGetWindowAttributes(xdpy_, win_, &win_attr)) {
        Log::error("XGetWindowAttributes failed!\n");
        return false;
    }

    char *pixel_data = reinterpret_cast<char *>(calloc(4 * width_ * height_, 1));

    ximage_ = XCreateImage(xdpy_, win_attr.visual, win_attr.depth,
                           ZPixmap, 0, pixel_data, width_, height_, 32, 0);
    if (!ximage_) {
        Log::error("XCreateImage failed!\n");
        return false;
    }

    return true;
}

static unsigned int
count_bits(unsigned long n)
{
    unsigned int c;
    for (c = 0; n; c++)
        n &= n - 1;
    return c;
}

bool
CompositeWindowPixman::ensure_pixman_image()
{
    if (pixman_image_)
        return true;

    unsigned int red_bits = count_bits(ximage_->red_mask);
    unsigned int green_bits = count_bits(ximage_->green_mask);
    unsigned int blue_bits = count_bits(ximage_->blue_mask);
    unsigned int alpha_bits = ximage_->depth - (red_bits + green_bits + blue_bits);

    pixman_format_code_t format =
        static_cast<pixman_format_code_t>(
                PIXMAN_FORMAT(ximage_->bits_per_pixel,
                              PIXMAN_TYPE_ARGB,
                              alpha_bits, red_bits, green_bits, blue_bits));

    Log::debug("Creating pixman image with bpp: %u format: a%ur%ug%ub%u\n",
               ximage_->bits_per_pixel, alpha_bits, red_bits, green_bits, blue_bits);

    if (!pixman_format_supported_source(format)) {
        Log::debug("Source pixmap format is not supported by pixman!");
    }
    else {
        pixman_image_ = pixman_image_create_bits(format, width_, height_,
                                                 reinterpret_cast<uint32_t *>(ximage_->data),
                                                 ximage_->bytes_per_line);
    }

    return pixman_image_ != 0;
}

void
CompositeWindowPixman::update_texture_from_pixmap(bool recreate)
{
    Profiler::ScopedSamples scoped_tfp(CompositeWindow::get_profiler_tfp_pair());

    if (recreate)
        release_resources();

    if (use_shm_) {
        if (!ensure_ximage_shm() || !ensure_pixman_image())
            return;

        XShmGetImage(xdpy_, pix_, ximage_, 0, 0, AllPlanes);
    }
    else {
        if (!ensure_ximage() || !ensure_pixman_image())
            return;

        XGetSubImage(xdpy_, pix_, 0, 0, width_, height_, AllPlanes, ZPixmap,
                     ximage_, 0, 0);
    }

}

bool
CompositeWindowPixman::update_texture()
{
    return false;
}
